/*
 * <sys/securebits.h>
 * Copyright (C) 2010	Serge Hallyn <serue@us.ibm.com>
 */

#ifndef _SYS_SECUREBITS_H
#define _SYS_SECUREBITS_H

#ifdef __cplusplus
extern "C" {
#endif

#ifndef __user
#define __user
#endif
#include <linux/securebits.h>

#ifdef __cplusplus
}
#endif

#endif /* _SYS_SECUREBITS_H */
